from typing_extensions import Any, Dict
from ripple_down_rules.datastructures.case import Case, create_case
from ripple_down_rules.helpers import general_rdr_classify
from . import physical_object_is_a_robot_output__scrdr as output__classifier

classifiers_dict = dict()
classifiers_dict['output_'] = output__classifier


def classify(case: Dict, **kwargs) -> Dict[str, Any]:
    if not isinstance(case, Case):
        case = create_case(case, max_recursion_idx=3)
    return general_rdr_classify(classifiers_dict, case, **kwargs)



import logging
from typing import Optional, List

from IPython.core.magic import magics_class, Magics, line_magic
from IPython.terminal.embed import InteractiveShellEmbed
from colorama import Fore, Style
from traitlets.config import Config

from ..datastructures.dataclasses import CaseQuery
from ..datastructures.enums import PromptFor
from .template_file_creator import TemplateFileCreator
from ..utils import contains_return_statement, extract_dependencies, encapsulate_code_lines_into_a_function


@magics_class
class MyMagics(Magics):

    def __init__(self, shell,
                 code_to_modify: Optional[str] = None,
                 prompt_for: Optional[PromptFor] = None,
                 case_query: Optional[CaseQuery] = None):
        super().__init__(shell)
        self.rule_editor = TemplateFileCreator(case_query, prompt_for=prompt_for, code_to_modify=code_to_modify)
        self.all_code_lines: Optional[List[str]] = None

    @line_magic
    def edit(self, line):
        self.rule_editor.edit()

    @line_magic
    def load(self, line):
        self.all_code_lines, updates = self.rule_editor.load(self.rule_editor.temp_file_path,
                                                             self.rule_editor.func_name,
                                                             self.rule_editor.print_func)
        self.shell.user_ns.update(updates)

    @line_magic
    def help(self, line):
        """
        Display help information for the Ipython shell.
        """
        help_text = f"""
Directly write python code in the shell, and then `{Fore.GREEN}return {Fore.RESET}output`. Or use 
the magic commands to write the code in a temporary file and edit it in PyCharm:
{Fore.MAGENTA}Usage: %edit{Style.RESET_ALL}
Opens a temporary file in PyCharm for editing a function (conclusion or conditions for case)
 that will be executed on the case object.
{Fore.MAGENTA}Usage: %load{Style.RESET_ALL}
Loads the function defined in the temporary file into the user namespace, that can then be used inside the
 Ipython shell. You can then do `{Fore.GREEN}return {Fore.RESET}function_name(case)`.
        """
        print(help_text)


class CustomInteractiveShell(InteractiveShellEmbed):
    def __init__(self, code_to_modify: Optional[str] = None,
                 prompt_for: Optional[PromptFor] = None,
                 case_query: Optional[CaseQuery] = None,
                 **kwargs):
        kwargs.update({'user_ns': case_query.scope})
        super().__init__(**kwargs)
        self.my_magics = MyMagics(self, code_to_modify=code_to_modify,
                                  prompt_for=prompt_for, case_query=case_query)
        self.register_magics(self.my_magics)
        self.all_lines = []

    def run_cell(self, raw_cell: str, **kwargs):
        """
        Override the run_cell method to capture return statements.
        """
        if contains_return_statement(raw_cell) and 'def ' not in raw_cell:
            if self.my_magics.rule_editor.func_name in raw_cell:
                self.all_lines = self.my_magics.all_code_lines
            self.all_lines.append(raw_cell)
            self.history_manager.store_inputs(line_num=self.execution_count, source=raw_cell)
            self.ask_exit()
            return None
        result = super().run_cell(raw_cell, **kwargs)
        if result.error_in_exec is None and result.error_before_exec is None:
            self.all_lines.append(raw_cell)
        return result


class IPythonShell:
    """
    Create an embedded Ipython shell that can be used to prompt the user for input.
    """

    def __init__(self, header: Optional[str] = None,
                 prompt_for: Optional[PromptFor] = None, case_query: Optional[CaseQuery] = None,
                 code_to_modify: Optional[str] = None):
        """
        Initialize the Ipython shell with the given scope and header.

        :param header: The header to display when the shell is started.
        :param prompt_for: The type of information to ask the user about.
        :param case_query: The case query which contains the case and the attribute to ask about.
        :param code_to_modify: The code to modify. If given, will be used as a start for user to modify.
        """
        self.header: str = header or ">>> Embedded Ipython Shell"
        self.case_query: Optional[CaseQuery] = case_query
        self.prompt_for: Optional[PromptFor] = prompt_for
        self.code_to_modify: Optional[str] = code_to_modify
        self.user_input: Optional[str] = None
        self.shell: CustomInteractiveShell = self._init_shell()
        self.all_code_lines: List[str] = []

    def _init_shell(self):
        """
        Initialize the Ipython shell with a custom configuration.
        """
        cfg = Config()
        shell = CustomInteractiveShell(config=cfg, banner1=self.header,
                                       code_to_modify=self.code_to_modify,
                                       prompt_for=self.prompt_for,
                                       case_query=self.case_query,
                                       )
        return shell

    def run(self):
        """
        Run the embedded shell.
        """
        while True:
            try:
                self.shell()
                self.update_user_input_from_code_lines()
                break
            except Exception as e:
                logging.error(e)
                print(f"{Fore.RED}ERROR:: {e}{Style.RESET_ALL}")

    def update_user_input_from_code_lines(self):
        """
        Update the user input from the code lines captured in the shell.
        """
        if self.shell.all_lines[-1] in ['quit', 'exit']:
            self.user_input = 'exit'
        elif self.shell.all_lines[0].replace('return', '').strip() == '':
            self.user_input = None
        else:
            self.all_code_lines = extract_dependencies(self.shell.all_lines)
            if len(self.all_code_lines) == 1 and self.all_code_lines[0].strip() == '':
                self.user_input = None
            else:
                self.user_input = encapsulate_code_lines_into_a_function(self.all_code_lines,
                                                       function_name=self.shell.my_magics.rule_editor.func_name,
                                                       function_signature=self.shell.my_magics.rule_editor.function_signature,
                                                       func_doc=self.shell.my_magics.rule_editor.func_doc,
                                                       case_query=self.case_query)



from __future__ import annotations
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.case import create_cases_from_dataframe
from ripple_down_rules.rdr import GeneralRDR
from ucimlrepo import fetch_ucirepo
from enum import Enum


class Species(str, Enum):
    """Enum for the species of the animals in the UCI Zoo dataset."""
    mammal = "mammal"
    bird = "bird"
    reptile = "reptile"
    fish = "fish"
    amphibian = "amphibian"
    insect = "insect"
    molusc = "molusc"

    @classmethod
    def from_str(cls, value: str) -> Species:
        return getattr(cls, value)


# fetch dataset
zoo = fetch_ucirepo(id=111)

# data (as pandas dataframes)
X = zoo.data.features
y = zoo.data.targets

# This is a utility that allows each row to be a Case instance,
# which simplifies access to column values using dot notation.
all_cases = create_cases_from_dataframe(X, name="Animal")

# The targets are the species of the animals
category_names = ["mammal", "bird", "reptile", "fish", "amphibian", "insect", "molusc"]
category_id_to_name = {i + 1: name for i, name in enumerate(category_names)}
targets = [Species.from_str(category_id_to_name[i]) for i in y.values.flatten()]

# Now that we are done with the data preparation, we can create and use the Ripple Down Rules classifier.
grdr = GeneralRDR()

# Fit the GRDR to the data
case_queries = [CaseQuery(case, 'species', type(target), True, _target=target)
                for case, target in zip(all_cases[:10], targets[:10])]
grdr.fit(case_queries)

# Classify a case
cat = grdr.classify(all_cases[50])['species']
assert cat == targets[50]


from ripple_down_rules.datastructures.case import Case, create_case
from typing_extensions import Optional
from types import NoneType
from .physical_object_is_a_robot_output__scrdr_defs import *


attribute_name = 'output_'
conclusion_type = (bool,)
mutually_exclusive = True


def classify(case: Dict, **kwargs) -> Optional[bool]:
    if not isinstance(case, Case):
        case = create_case(case, max_recursion_idx=3)

    if conditions_226969243620390858682731042391766665817(case):
        return conclusion_226969243620390858682731042391766665817(case)
    else:
        return None



from __future__ import annotations

import logging
import re
from abc import ABC, abstractmethod
from pathlib import Path
from types import NoneType
from uuid import uuid4

from anytree import NodeMixin
from sqlalchemy.orm import DeclarativeBase as SQLTable
from typing_extensions import List, Optional, Self, Union, Dict, Any, Tuple, Type, Set

from .datastructures.callable_expression import CallableExpression
from .datastructures.case import Case
from .datastructures.dataclasses import CaseFactoryMetaData, CaseQuery
from .datastructures.enums import RDREdge, Stop
from .utils import SubclassJSONSerializer, conclusion_to_json, get_full_class_name, get_imports_from_types


class Rule(NodeMixin, SubclassJSONSerializer, ABC):
    fired: Optional[bool] = None
    """
    Whether the rule has fired or not.
    """

    def __init__(self, conditions: Optional[CallableExpression] = None,
                 conclusion: Optional[CallableExpression] = None,
                 parent: Optional[Rule] = None,
                 corner_case: Optional[Union[Case, SQLTable]] = None,
                 weight: Optional[str] = None,
                 conclusion_name: Optional[str] = None,
                 uid: Optional[str] = None,
                 corner_case_metadata: Optional[CaseFactoryMetaData] = None):
        """
        A rule in the ripple down rules classifier.

        :param conditions: The conditions of the rule.
        :param conclusion: The conclusion of the rule when the conditions are met.
        :param parent: The parent rule of this rule.
        :param corner_case: The corner case that this rule is based on/created from.
        :param weight: The weight of the rule, which is the type of edge connecting the rule to its parent.
        :param conclusion_name: The name of the conclusion of the rule.
        :param uid: The unique id of the rule.
        :param corner_case_metadata: Metadata about the corner case, such as the factory that created it or the
         scenario it is based on.
        """
        super(Rule, self).__init__()
        self.conclusion = conclusion
        self.corner_case = corner_case
        self.corner_case_metadata: Optional[CaseFactoryMetaData] = corner_case_metadata
        self.parent = parent
        self.weight: Optional[str] = weight
        self.conditions = conditions if conditions else None
        self.conclusion_name: Optional[str] = conclusion_name
        self.json_serialization: Optional[Dict[str, Any]] = None
        self._name: Optional[str] = None
        # generate a unique id for the rule using uuid4
        self.uid: str = uid if uid else str(uuid4().int)
        self.evaluated: bool = False
        self._user_defined_name: Optional[str] = None

    @property
    def user_defined_name(self) -> Optional[str]:
        """
        Get the user defined name of the rule, if it exists.
        """
        if self._user_defined_name is None:
            if self.conditions and self.conditions.user_input and "def " in self.conditions.user_input:
                # If the conditions have a user input, use it as the name
                self._user_defined_name = self.conditions.user_input.split('(')[0].replace('def ', '').strip()
            else:
                self._user_defined_name = f"Rule_{self.uid}"
        return self._user_defined_name

    @classmethod
    def from_case_query(cls, case_query: CaseQuery) -> Rule:
        """
        Create a SingleClassRule from a CaseQuery.

        :param case_query: The CaseQuery to create the rule from.
        :return: A SingleClassRule instance.
        """
        corner_case_metadata = CaseFactoryMetaData.from_case_query(case_query)
        return cls(conditions=case_query.conditions, conclusion=case_query.target,
                   corner_case=case_query.case, parent=None,
                   corner_case_metadata=corner_case_metadata,
                   conclusion_name=case_query.attribute_name)

    def _post_detach(self, parent):
        """
        Called after this node is detached from the tree, useful when drawing the tree.

        :param parent: The parent node from which this node was detached.
        """
        self.weight = None

    def __call__(self, x: Case) -> Self:
        return self.evaluate(x)

    def evaluate(self, x: Case) -> Rule:
        """
        Check if the rule or its refinement or its alternative match the case,
        by checking if the conditions are met, then return the rule that matches the case.

        :param x: The case to evaluate the rule on.
        :return: The rule that fired or the last evaluated rule if no rule fired.
        """
        if self.root is self:
            for descendant in self.descendants:
                descendant.evaluated = False
        self.evaluated = True
        if not self.conditions:
            raise ValueError("Rule has no conditions")
        self.fired = self.conditions(x)
        return self.evaluate_next_rule(x)

    @abstractmethod
    def evaluate_next_rule(self, x: Case):
        """
        Evaluate the next rule after this rule is evaluated.
        """
        pass

    def write_corner_case_as_source_code(self, cases_file: str, package_name: Optional[str] = None) -> None:
        """
        Write the source code representation of the corner case of the rule to a file.

        :param cases_file: The file to write the corner case to.
        :param package_name: The package name to use for relative imports.
        """
        if self.corner_case_metadata is None:
            return
        with open(cases_file, 'a') as f:
            f.write(f"corner_case_{self.uid} = {self.corner_case_metadata}" + "\n\n\n")

    def get_corner_case_types_to_import(self) -> Set[Type]:
        """
        Get the types that need to be imported for the corner case of the rule.
        """
        if self.corner_case_metadata is None:
            return
        types_to_import = set()
        if self.corner_case_metadata.factory_method is not None:
            types_to_import.add(self.corner_case_metadata.factory_method)
        if self.corner_case_metadata.scenario is not None:
            types_to_import.add(self.corner_case_metadata.scenario)
        if self.corner_case_metadata.case_conf is not None:
            types_to_import.add(self.corner_case_metadata.case_conf)
        return types_to_import

    def write_conclusion_as_source_code(self, parent_indent: str = "", defs_file: Optional[str] = None) -> str:
        """
        Get the source code representation of the conclusion of the rule.

        :param parent_indent: The indentation of the parent rule.
        :param defs_file: The file to write the conclusion to if it is a definition.
        :return: The source code representation of the conclusion of the rule.
        """
        if self.conclusion.user_input is not None:
            conclusion = self.conclusion.user_input
        else:
            conclusion = self.conclusion.conclusion
        conclusion_func, conclusion_func_call = self.get_conclusion_as_source_code(conclusion,
                                                                                   parent_indent=parent_indent)
        if conclusion_func is not None:
            with open(defs_file, 'a') as f:
                f.write(conclusion_func.strip() + "\n\n\n")
        return conclusion_func_call

    def get_conclusion_as_source_code(self, conclusion: Any, parent_indent: str = "") -> Tuple[Optional[str], str]:
        """
        Convert the conclusion of a rule to source code.

        :param conclusion: The conclusion to convert to source code.
        :param parent_indent: The indentation of the parent rule.
        :return: The source code of the conclusion as a tuple of strings, one for the function and one for the call.
        """
        if "def " in conclusion:
            # This means the conclusion is a definition that should be written and then called
            conclusion_lines = conclusion.split('\n')
            # use regex to replace the function name
            new_function_name = f"def conclusion_{self.uid}"
            conclusion_lines[0] = re.sub(r"def (\w+)", new_function_name, conclusion_lines[0])
            # add type hint
            if len(self.conclusion.conclusion_type) == 1:
                hint = self.conclusion.conclusion_type[0].__name__
            else:
                if (all(t in self.conclusion.conclusion_type for t in [list, set])
                        and len(self.conclusion.conclusion_type) > 2):
                    type_names = [t.__name__ for t in self.conclusion.conclusion_type if t not in [list, set]]
                    hint = f"List[{', '.join(type_names)}]"
                else:
                    if NoneType in self.conclusion.conclusion_type:
                        type_names = [t.__name__ for t in self.conclusion.conclusion_type if t is not NoneType]
                        hint = f"Optional[{', '.join(type_names)}]"
                    else:
                        type_names = [t.__name__ for t in self.conclusion.conclusion_type]
                        hint = f"Union[{', '.join(type_names)}]"
            conclusion_lines[0] = conclusion_lines[0].replace("):", f") -> {hint}:")
            func_call = f"{parent_indent}    return {new_function_name.replace('def ', '')}(case)\n"
            return "\n".join(conclusion_lines).strip(' '), func_call
        else:
            raise ValueError(f"Conclusion format is not valid, it should contain a function definition."
                             f" Instead got:\n{conclusion}\n")

    def write_condition_as_source_code(self, parent_indent: str = "", defs_file: Optional[str] = None) -> str:
        """
        Get the source code representation of the conditions of the rule.

        :param parent_indent: The indentation of the parent rule.
        :param defs_file: The file to write the conditions to if they are a definition.
        """
        if_clause = self._if_statement_source_code_clause()
        if "def " in self.conditions.user_input:
            if defs_file is None:
                raise ValueError("Cannot write conditions to source code as definitions python file was not given.")
            # This means the conditions are a definition that should be written and then called
            conditions_lines = self.conditions.user_input.split('\n')
            # use regex to replace the function name
            new_function_name = f"def conditions_{self.uid}"
            conditions_lines[0] = re.sub(r"def (\w+)", new_function_name, conditions_lines[0])
            # add type hint
            conditions_lines[0] = conditions_lines[0].replace('):', ') -> bool:')
            def_code = "\n".join(conditions_lines)
            with open(defs_file, 'a') as f:
                f.write(def_code.strip() + "\n\n\n")
            return f"\n{parent_indent}{if_clause} {new_function_name.replace('def ', '')}(case):\n"
        else:
            raise ValueError(f"Conditions format is not valid, it should contain a function definition"
                             f" Instead got:\n{self.conditions.user_input}\n")

    @abstractmethod
    def _if_statement_source_code_clause(self) -> str:
        pass

    def _to_json(self) -> Dict[str, Any]:
        try:
            corner_case = SubclassJSONSerializer.to_json_static(self.corner_case) if self.corner_case else None
        except Exception as e:
            logging.debug("Failed to serialize corner case to json, setting it to None. Error: %s", e)
            corner_case = None
        json_serialization = {"_type": get_full_class_name(type(self)),
                              "conditions": self.conditions.to_json(),
                              "conclusion": conclusion_to_json(self.conclusion),
                              "parent": self.parent.json_serialization if self.parent else None,
                              "corner_case": corner_case,
                              "conclusion_name": self.conclusion_name,
                              "weight": self.weight,
                              "uid": self.uid}
        return json_serialization

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Rule:
        try:
            corner_case = Case.from_json(data["corner_case"])
        except Exception as e:
            logging.debug("Failed to load corner case from json, setting it to None.")
            corner_case = None
        loaded_rule = cls(conditions=CallableExpression.from_json(data["conditions"]),
                          conclusion=CallableExpression.from_json(data["conclusion"]),
                          parent=cls.from_json(data["parent"]),
                          corner_case=corner_case,
                          conclusion_name=data["conclusion_name"],
                          weight=data["weight"],
                          uid=data["uid"])
        return loaded_rule

    @property
    def name(self):
        """
        Get the name of the rule, which is the conditions and the conclusion.
        """
        return self._name if self._name is not None else self.__str__()

    @name.setter
    def name(self, new_name: str):
        """
        Set the name of the rule.
        """
        self._name = new_name

    @property
    def semantic_condition_name(self) -> Optional[str]:
        """
        Get the name of the conditions of the rule, which is the user input of the conditions.
        """
        return self.expression_name(self.conditions)

    @property
    def semantic_conclusion_name(self) -> Optional[str]:
        """
        Get the name of the conclusion of the rule, which is the user input of the conclusion.
        """
        return self.expression_name(self.conclusion)

    @staticmethod
    def expression_name(expression: CallableExpression) -> str:
        """
        Get the name of the expression, which is the user input of the expression if it exists,
        otherwise it is the conclusion or conditions of the rule.
        """
        if expression.user_defined_name is not None:
            return expression.user_defined_name.strip()
        elif expression.user_input and "def " in expression.user_input:
            return expression.user_input.split('(')[0].replace('def ', '').strip()
        elif expression.user_input:
            return expression.user_input.strip()
        else:
            return str(expression)

    def __str__(self, sep="\n"):
        """
        Get the string representation of the rule, which is the conditions and the conclusion.
        """
        return f"{self.semantic_condition_name}{sep}=> {self.semantic_conclusion_name}"

    def __repr__(self):
        return self.__str__()


class HasAlternativeRule:
    """
    A mixin class for rules that have an alternative rule.
    """
    _alternative: Optional[Rule] = None
    """
    The alternative rule of the rule, which is evaluated when the rule doesn't fire.
    """
    furthest_alternative: Optional[List[Rule]] = None
    """
    The furthest alternative rule of the rule, which is the last alternative rule in the chain of alternative rules.
    """
    all_alternatives: Optional[List[Rule]] = None
    """
    All alternative rules of the rule, which is all the alternative rules in the chain of alternative rules.
    """

    @property
    def alternative(self) -> Optional[Rule]:
        return self._alternative

    @alternative.setter
    def alternative(self, new_rule: Rule):
        """
        Set the alternative rule of the rule. It is important that no rules should be retracted or changed,
        only new rules should be added.
        """
        if new_rule is None:
            return
        if self.furthest_alternative:
            self.furthest_alternative[-1].alternative = new_rule
        else:
            new_rule.parent = self
            new_rule.weight = RDREdge.Alternative.value if not new_rule.weight else new_rule.weight
            self._alternative = new_rule
        self.furthest_alternative = [new_rule]


class HasRefinementRule:
    _refinement: Optional[HasAlternativeRule] = None
    """
    The refinement rule of the rule, which is evaluated when the rule fires.
    """

    @property
    def refinement(self) -> Optional[Rule]:
        return self._refinement

    @refinement.setter
    def refinement(self, new_rule: Rule):
        """
        Set the refinement rule of the rule. It is important that no rules should be retracted or changed,
        only new rules should be added.
        """
        if new_rule is None:
            return
        new_rule.top_rule = self
        if self.refinement:
            self.refinement.alternative = new_rule
        else:
            new_rule.parent = self
            new_rule.weight = RDREdge.Refinement.value
            self._refinement = new_rule


class SingleClassRule(Rule, HasAlternativeRule, HasRefinementRule):
    """
    A rule in the SingleClassRDR classifier, it can have a refinement or an alternative rule or both.
    """

    def evaluate_next_rule(self, x: Case) -> SingleClassRule:
        if self.fired:
            returned_rule = self.refinement(x) if self.refinement else self
        else:
            returned_rule = self.alternative(x) if self.alternative else self
        return returned_rule if returned_rule.fired else self

    def fit_rule(self, case_query: CaseQuery):
        corner_case_metadata = CaseFactoryMetaData.from_case_query(case_query)
        new_rule = SingleClassRule(case_query.conditions, case_query.target,
                                   corner_case=case_query.case, parent=self,
                                   corner_case_metadata=corner_case_metadata,
                                   )
        if self.fired:
            self.refinement = new_rule
        else:
            self.alternative = new_rule

    def _to_json(self) -> Dict[str, Any]:
        self.json_serialization = {**super(SingleClassRule, self)._to_json(),
                                   "refinement": self.refinement.to_json() if self.refinement is not None else None,
                                   "alternative": self.alternative.to_json() if self.alternative is not None else None}
        return self.json_serialization

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> SingleClassRule:
        loaded_rule = super(SingleClassRule, cls)._from_json(data)
        loaded_rule.refinement = SingleClassRule.from_json(data["refinement"])
        loaded_rule.alternative = SingleClassRule.from_json(data["alternative"])
        return loaded_rule

    def _if_statement_source_code_clause(self) -> str:
        return "elif" if self.weight == RDREdge.Alternative.value else "if"


class MultiClassStopRule(Rule, HasAlternativeRule):
    """
    A rule in the MultiClassRDR classifier, it can have an alternative rule and a top rule,
    the conclusion of the rule is a Stop category meant to stop the parent conclusion from being made.
    """
    top_rule: Optional[MultiClassTopRule] = None
    """
    The top rule of the rule, which is the nearest ancestor that fired and this rule is a refinement of.
    """

    def __init__(self, *args, **kwargs):
        super(MultiClassStopRule, self).__init__(*args, **kwargs)
        self.conclusion = CallableExpression(conclusion_type=(Stop,), conclusion=Stop.stop)

    def evaluate_next_rule(self, x: Case) -> Optional[Union[MultiClassStopRule, MultiClassTopRule]]:
        if self.fired:
            self.top_rule.fired = False
            return self.top_rule.alternative
        elif self.alternative:
            return self.alternative(x)
        else:
            return self.top_rule.alternative

    def _to_json(self) -> Dict[str, Any]:
        self.json_serialization = {**Rule._to_json(self),
                                   "alternative": self.alternative.to_json() if self.alternative is not None else None}
        return self.json_serialization

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> MultiClassStopRule:
        loaded_rule = super(MultiClassStopRule, cls)._from_json(data)
        # The following is done to prevent re-initialization of the top rule,
        # so the top rule that is already initialized is passed in the data instead of its json serialization.
        loaded_rule.top_rule = data['top_rule']
        if data['alternative'] is not None:
            data['alternative']['top_rule'] = data['top_rule']
        loaded_rule.alternative = MultiClassStopRule.from_json(data["alternative"])
        return loaded_rule

    def get_conclusion_as_source_code(self, conclusion: Any, parent_indent: str = "") -> Tuple[None, str]:
        return None, f"{parent_indent}{' ' * 4}pass\n"

    def _if_statement_source_code_clause(self) -> str:
        return "elif" if self.weight == RDREdge.Alternative.value else "if"


class MultiClassTopRule(Rule, HasRefinementRule, HasAlternativeRule):
    """
    A rule in the MultiClassRDR classifier, it can have a refinement and a next rule.
    """

    def __init__(self, *args, **kwargs):
        super(MultiClassTopRule, self).__init__(*args, **kwargs)
        self.weight = RDREdge.Next.value

    def evaluate_next_rule(self, x: Case) -> Optional[Union[MultiClassStopRule, MultiClassTopRule]]:
        if self.fired and self.refinement:
            return self.refinement(x)
        elif self.alternative:  # Here alternative refers to next rule in MultiClassRDR
            return self.alternative

    def fit_rule(self, case_query: CaseQuery):
        if self.fired and case_query.target != self.conclusion:
            self.refinement = MultiClassStopRule(case_query.conditions, corner_case=case_query.case, parent=self)
        elif not self.fired:
            self.alternative = MultiClassTopRule(case_query.conditions, case_query.target,
                                                 corner_case=case_query.case, parent=self)

    def _to_json(self) -> Dict[str, Any]:
        self.json_serialization = {**Rule._to_json(self),
                                   "refinement": self.refinement.to_json() if self.refinement is not None else None,
                                   "alternative": self.alternative.to_json() if self.alternative is not None else None}
        return self.json_serialization

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> MultiClassTopRule:
        loaded_rule = super(MultiClassTopRule, cls)._from_json(data)
        # The following is done to prevent re-initialization of the top rule,
        # so the top rule that is already initialized is passed in the data instead of its json serialization.
        if data['refinement'] is not None:
            data['refinement']['top_rule'] = loaded_rule
        loaded_rule.refinement = MultiClassStopRule.from_json(data["refinement"])
        loaded_rule.alternative = MultiClassTopRule.from_json(data["alternative"])
        return loaded_rule

    def get_conclusion_as_source_code(self, conclusion: Any, parent_indent: str = "") -> Tuple[str, str]:
        func, func_call = super().get_conclusion_as_source_code(str(conclusion), parent_indent=parent_indent)
        conclusion_str = func_call.replace("return ", "").strip()

        statement = f"{parent_indent}    conclusions.update(make_set({conclusion_str}))\n"
        if self.alternative is None:
            statement += f"{parent_indent}return conclusions\n"
        return func, statement

    def _if_statement_source_code_clause(self) -> str:
        return "if"



from ripple_down_rules.utils import make_set
from ripple_down_rules.datastructures.case import Case, create_case
from typing_extensions import Optional, Set
from .physical_object_select_objects_that_are_parts_of_robot_output__mcrdr_defs import *


attribute_name = 'output_'
conclusion_type = (set, PhysicalObject, list,)
mutually_exclusive = False


def classify(case: Dict, **kwargs) -> Set[PhysicalObject]:
    if not isinstance(case, Case):
        case = create_case(case, max_recursion_idx=3)
    conclusions = set()

    if conditions_164855806603893754507167918997373216146(case):
        conclusions.update(make_set(conclusion_164855806603893754507167918997373216146(case)))
    return conclusions



from __future__ import annotations

import os
import pickle
from dataclasses import dataclass, field

import sqlalchemy
from sqlalchemy import ForeignKey
from sqlalchemy.orm import MappedAsDataclass, Mapped, mapped_column, relationship, MappedColumn
from typing_extensions import Tuple, List, Set, Optional, Self
from ucimlrepo import fetch_ucirepo

from ripple_down_rules.datastructures.case import Case, create_cases_from_dataframe
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.rdr_decorators import RDRDecorator


def load_cached_dataset(cache_file):
    """Loads the dataset from cache if it exists."""
    dataset = {}
    if '.pkl' not in cache_file:
        cache_file += ".pkl"
    for key in ["features", "targets", "ids"]:
        part_file = cache_file.replace(".pkl", f"_{key}.pkl")
        if not os.path.exists(part_file):
            return None
        with open(part_file, "rb") as f:
            dataset[key] = pickle.load(f)
    return dataset


def save_dataset_to_cache(dataset, cache_file):
    """Saves only essential parts of the dataset to cache."""
    dataset_to_cache = {
        "features": dataset.data.features,
        "targets": dataset.data.targets,
        "ids": dataset.data.ids,
    }

    for key, value in dataset_to_cache.items():
        with open(cache_file.replace(".pkl", f"_{key}.pkl"), "wb") as f:
            pickle.dump(dataset_to_cache[key], f)
    print("Dataset cached successfully.")


def get_dataset(dataset_id, cache_file: Optional[str] = None):
    """Fetches dataset from cache or downloads it if not available."""
    if cache_file is not None:
        if not cache_file.endswith(".pkl"):
            cache_file += ".pkl"
    dataset = load_cached_dataset(cache_file) if cache_file else None
    if dataset is None:
        print("Downloading dataset...")
        dataset = fetch_ucirepo(id=dataset_id)

        # Check if dataset is valid before caching
        if dataset is None or not hasattr(dataset, "data"):
            print("Error: Failed to fetch dataset.")
            return None

        if cache_file:
            save_dataset_to_cache(dataset, cache_file)

        dataset = {
            "features": dataset.data.features,
            "targets": dataset.data.targets,
            "ids": dataset.data.ids,
        }

    return dataset


def load_zoo_dataset(cache_file: Optional[str] = None) -> Tuple[List[Case], List[Species]]:
    """
    Load the zoo dataset.

    :param cache_file: the cache file to store the dataset or load it from.
    :return: all cases and targets.
    """
    # fetch dataset
    zoo = get_dataset(111, cache_file)

    # data (as pandas dataframes)
    X = zoo['features']
    y = zoo['targets']
    # get ids as list of strings
    ids = zoo['ids'].values.flatten()
    all_cases = create_cases_from_dataframe(X, name="Animal")

    category_names = ["mammal", "bird", "reptile", "fish", "amphibian", "insect", "molusc"]
    category_id_to_name = {i + 1: name for i, name in enumerate(category_names)}
    # targets = [getattr(SpeciesCol, category_id_to_name[i]) for i in y.values.flatten()]
    targets = [Species.from_str(category_id_to_name[i]) for i in y.values.flatten()]
    return all_cases, targets


def load_zoo_cases(cache_file: Optional[str] = None) -> List[Case]:
    """
    Load the zoo dataset cases.

    :param cache_file: the cache file to store the dataset or load it from.
    :return: all cases.
    """
    all_cases, _ = load_zoo_dataset(cache_file)
    return all_cases


class Species(Category):
    mammal = "mammal"
    bird = "bird"
    reptile = "reptile"
    fish = "fish"
    amphibian = "amphibian"
    insect = "insect"
    molusc = "molusc"


class Habitat(Category):
    """
    A habitat category is a category that represents the habitat of an animal.
    """
    land = "land"
    water = "water"
    air = "air"


class PhysicalObject:
    """
    A physical object is an object that can be contained in a container.
    """
    _rdr_json_dir: str = os.path.join(os.path.dirname(__file__), "test_results")
    """
    The directory where the RDR serialized JSON files are stored.
    """
    _is_a_robot_rdr: RDRDecorator = RDRDecorator(_rdr_json_dir, (bool,), True,
                                                 package_name="test")
    """
    The RDR decorator that is used to determine if the object is a robot or not.
    """
    _select_parts_rdr: RDRDecorator = RDRDecorator(_rdr_json_dir, (Self,), False,
                                                   package_name="test")
    """
    The RDR decorator that is used to determine if the object is a robot or not.
    """

    def __init__(self, name: str, contained_objects: Optional[List[PhysicalObject]] = None):
        self.name: str = name
        self._contained_objects: List[PhysicalObject] = contained_objects or []

    @property
    def contained_objects(self) -> List[PhysicalObject]:
        return self._contained_objects

    @contained_objects.setter
    def contained_objects(self, value: List[PhysicalObject]):
        self._contained_objects = value

    @_is_a_robot_rdr.decorator
    def is_a_robot(self) -> bool:
        pass

    @_select_parts_rdr.decorator
    def select_objects_that_are_parts_of_robot(self, objects: List[PhysicalObject], robot: Robot) -> List[PhysicalObject]:
        pass

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


class Part(PhysicalObject):
    ...


class Robot(PhysicalObject):

    def __init__(self, name: str, parts: Optional[List[Part]] = None):
        super().__init__(name)
        self.parts: List[Part] = parts if parts else []


class Base(sqlalchemy.orm.DeclarativeBase):
    pass


class HabitatTable(MappedAsDataclass, Base):
    __tablename__ = "Habitat"

    id: Mapped[int] = mapped_column(init=False, primary_key=True, autoincrement=True)
    habitat: Mapped[Habitat]
    animal_id: MappedColumn = mapped_column(ForeignKey("Animal.id"), init=False)

    def __hash__(self):
        return hash(self.habitat)

    def __str__(self):
        return f"{HabitatTable.__name__}({Habitat.__name__}.{self.habitat.name})"

    def __repr__(self):
        return self.__str__()


class MappedAnimal(MappedAsDataclass, Base):
    __tablename__ = "Animal"

    id: Mapped[int] = mapped_column(init=False, primary_key=True, autoincrement=True)
    name: Mapped[str]
    hair: Mapped[bool]
    feathers: Mapped[bool]
    eggs: Mapped[bool]
    milk: Mapped[bool]
    airborne: Mapped[bool]
    aquatic: Mapped[bool]
    predator: Mapped[bool]
    toothed: Mapped[bool]
    backbone: Mapped[bool]
    breathes: Mapped[bool]
    venomous: Mapped[bool]
    fins: Mapped[bool]
    legs: Mapped[int]
    tail: Mapped[bool]
    domestic: Mapped[bool]
    catsize: Mapped[bool]
    species: Mapped[Species] = mapped_column(nullable=True)

    habitats: Mapped[Set[HabitatTable]] = relationship(default_factory=set)


@dataclass
class WorldEntity:
    world: Optional[World] = field(default=None, kw_only=True, repr=False, hash=False)


@dataclass(unsafe_hash=True)
class Body(WorldEntity):
    name: str


@dataclass(unsafe_hash=True)
class Handle(Body):
    ...


@dataclass(unsafe_hash=True)
class Container(Body):
    ...


@dataclass(unsafe_hash=True)
class Connection(WorldEntity):
    parent: Body
    child: Body


@dataclass(unsafe_hash=True)
class FixedConnection(Connection):
    ...


@dataclass(unsafe_hash=True)
class PrismaticConnection(Connection):
    ...


@dataclass
class World:
    id: int = 0
    bodies: List[Body] = field(default_factory=list)
    connections: List[Connection] = field(default_factory=list)
    views: List[View] = field(default_factory=list, repr=False)

    def __eq__(self, other):
        if not isinstance(other, World):
            return False
        return self.id == other.id


@dataclass(unsafe_hash=True)
class View(WorldEntity):
    ...


@dataclass(unsafe_hash=True)
class Drawer(View):
    handle: Handle
    container: Container
    correct: Optional[bool] = None


@dataclass
class Cabinet(View):
    container: Container
    drawers: List[Drawer] = field(default_factory=list)

    def __hash__(self):
        return hash((self.__class__.__name__, self.container))





from __future__ import annotations

from dataclasses import dataclass, field

from omegaconf import MISSING
from typing_extensions import List, TYPE_CHECKING

from ripple_down_rules.datastructures.dataclasses import CaseConf

if TYPE_CHECKING:
    pass


@dataclass
class BodyConf:
    name: str = MISSING


@dataclass
class HandleConf(BodyConf):
    ...


@dataclass
class ContainerConf(BodyConf):
    ...


@dataclass
class Connection:
    parent: BodyConf = MISSING
    child: BodyConf = MISSING


@dataclass
class FixedConnectionConf(Connection):
    pass


@dataclass
class PrismaticConnectionConf(Connection):
    pass


@dataclass
class WorldConf(CaseConf):
    bodies: List[BodyConf] = field(default_factory=list)
    connections: List[Connection] = field(default_factory=list)



import unittest
from .datasets import PhysicalObject, Part, Robot


class RDRDecoratorsTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.robot = Robot("Robot1", [Part("Part1"), Part("Part2")])
        cls.physical_object = PhysicalObject("Object1", [cls.robot])
        cls.part_3 = Part("Part3")

    def test_is_a_robot(self):
        self.physical_object._is_a_robot_rdr.fit = False
        self.assertFalse(self.physical_object.is_a_robot())
        self.assertTrue(self.robot.is_a_robot())
        self.physical_object._is_a_robot_rdr.save()

    def test_select_objects_that_are_parts_of_robots(self):
        self.physical_object._select_parts_rdr.fit = False
        selected_parts = self.physical_object.select_objects_that_are_parts_of_robot([self.part_3, self.robot,
                                                                                      *self.robot.parts], self.robot)
        for part in selected_parts:
            self.assertIn(part, self.robot.parts)
        self.assertNotIn(self.part_3, selected_parts)
        self.assertNotIn(self.robot, selected_parts)
        self.physical_object._select_parts_rdr.save()



from __future__ import annotations

from os.path import dirname

from ripple_down_rules.utils import draw_tree, render_tree

try:
    from PyQt6.QtWidgets import QApplication
    from ripple_down_rules.user_interface.gui import RDRCaseViewer
except ImportError as e:
    QApplication = None
    RDRCaseViewer = None

from ..datasets import *
from ripple_down_rules.helpers import is_matching
from ripple_down_rules.rdr import GeneralRDR



def test_drawer_cabinet_rdr(drawer_cabinet_rdr):
    pass


def test_save_and_load_drawer_cabinet_rdr(handles_and_containers_world, drawer_cabinet_rdr):
    world = handles_and_containers_world
    filename = os.path.join(dirname(__file__), "../test_results/world_drawer_cabinet_rdr")
    model_name = drawer_cabinet_rdr.save(filename)
    loaded_rdr = GeneralRDR.load(filename, model_name=model_name)
    assert drawer_cabinet_rdr.classify(world) == loaded_rdr.classify(world)
    assert world.bodies == loaded_rdr.start_rules[0].corner_case.bodies


def test_draw_evaluated_tree_for_drawer_cabinet_rdr(handles_and_containers_world, drawer_cabinet_rdr):
    world = handles_and_containers_world
    filename = os.path.join(dirname(__file__), "../test_results/world_drawer_cabinet_rdr")
    loaded_rdr = GeneralRDR.load(filename, model_name="world_rdr")
    loaded_rdr.classify(world)
    loaded_rdr.render_evaluated_rule_tree("drawer_cabinet_rdr_evaluated_tree")
    assert True  # If no exception is raised, the test passes


def test_write_drawer_cabinet_rdr_to_python_file(drawer_cabinet_rdr, handles_and_containers_world):
    rdrs_dir = os.path.join(dirname(__file__), "../test_generated_rdrs/view_rdr")
    os.makedirs(rdrs_dir, exist_ok=True)
    drawer_cabinet_rdr._write_to_python(rdrs_dir)
    loaded_rdr_classifier = drawer_cabinet_rdr.get_rdr_classifier_from_python_file(rdrs_dir)
    found_views = loaded_rdr_classifier(handles_and_containers_world)
    assert len([v for v in found_views["views"] if isinstance(v, Drawer)]) == 1
    assert len([v for v in found_views["views"] if isinstance(v, Cabinet)]) == 1
    assert len(found_views["views"]) == 2


def test_drawer_rdr(correct_drawer_rdr):
    pass


def test_write_drawer_rdr_to_python_file(correct_drawer_rdr, drawer_case_queries):
    rdrs_dir = os.path.join(dirname(__file__), "../test_generated_rdrs/drawer_rdr")
    os.makedirs(rdrs_dir, exist_ok=True)
    correct_drawer_rdr._write_to_python(rdrs_dir)
    loaded_rdr_classifier = correct_drawer_rdr.get_rdr_classifier_from_python_file(rdrs_dir)
    for case_query in drawer_case_queries:
        assert is_matching(loaded_rdr_classifier, case_query)



import ast
import logging
from _ast import AST

try:
    from PyQt6.QtWidgets import QApplication
    from .gui import RDRCaseViewer
except ImportError:
    QApplication = None
    RDRCaseViewer = None

from colorama import Fore, Style
from pygments import highlight
from pygments.formatters.terminal import TerminalFormatter
from pygments.lexers.python import PythonLexer
from typing_extensions import Optional, Tuple

from ..datastructures.callable_expression import CallableExpression, parse_string_to_expression
from ..datastructures.dataclasses import CaseQuery
from ..datastructures.enums import PromptFor
from .ipython_custom_shell import IPythonShell
from ..utils import make_list
from threading import RLock


class UserPrompt:
    """
    A class to handle user prompts for the RDR.
    """
    shell_lock: RLock = RLock()  # To ensure that only one thread can access the shell at a time

    def __init__(self, viewer: Optional[RDRCaseViewer] = None):
        """
        Initialize the UserPrompt class.

        :param viewer: The RDRCaseViewer instance to use for prompting the user.
        """
        self.viewer = viewer
        self.print_func = print if viewer is None else viewer.print


    def prompt_user_for_expression(self, case_query: CaseQuery, prompt_for: PromptFor, prompt_str: Optional[str] = None) \
            -> Tuple[Optional[str], Optional[CallableExpression]]:
        """
        Prompt the user for an executable python expression to the given case query.

        :param case_query: The case query to prompt the user for.
        :param prompt_for: The type of information ask user about.
        :param prompt_str: The prompt string to display to the user.
        :return: A callable expression that takes a case and executes user expression on it.
        """
        prev_user_input: Optional[str] = None
        callable_expression: Optional[CallableExpression] = None
        while True:
            with self.shell_lock:
                user_input, expression_tree = self.prompt_user_about_case(case_query, prompt_for, prompt_str, code_to_modify=prev_user_input)
            if user_input is None:
                if prompt_for == PromptFor.Conclusion:
                    self.print_func(f"{Fore.YELLOW}No conclusion provided. Exiting.{Style.RESET_ALL}")
                    return None, None
                else:
                    self.print_func(f"{Fore.RED}Conditions must be provided. Please try again.{Style.RESET_ALL}")
                    continue
            elif user_input == "exit":
                self.print_func(f"{Fore.YELLOW}Exiting.{Style.RESET_ALL}")
                return user_input, None
                
            prev_user_input = '\n'.join(user_input.split('\n')[2:-1])
            conclusion_type = bool if prompt_for == PromptFor.Conditions else case_query.attribute_type
            callable_expression = CallableExpression(user_input, conclusion_type, expression_tree=expression_tree,
                                                     scope=case_query.scope,
                                                      mutually_exclusive=case_query.mutually_exclusive)
            try:
                result = callable_expression(case_query.case)
                if len(make_list(result)) == 0:
                    self.print_func(f"{Fore.YELLOW}The given expression gave an empty result for case {case_query.name}."
                          f" Please modify!{Style.RESET_ALL}")
                    continue
                break
            except Exception as e:
                logging.error(e)
                self.print_func(f"{Fore.RED}{e}{Style.RESET_ALL}")
        return user_input, callable_expression


    def prompt_user_about_case(self, case_query: CaseQuery, prompt_for: PromptFor,
                               prompt_str: Optional[str] = None,
                               code_to_modify: Optional[str] = None) -> Tuple[Optional[str], Optional[AST]]:
        """
        Prompt the user for input.

        :param case_query: The case query to prompt the user for.
        :param prompt_for: The type of information the user should provide for the given case.
        :param prompt_str: The prompt string to display to the user.
        :param code_to_modify: The code to modify. If given will be used as a start for user to modify.
        :return: The user input, and the executable expression that was parsed from the user input.
        """
        self.print_func("Entered shell")
        prompt_str = f"{Fore.WHITE}{prompt_str}" if prompt_str is not None else ''
        if prompt_for == PromptFor.Conclusion:
            prompt_str += f"\n{Fore.MAGENTA}Give possible value(s) for:"
        else:
            prompt_str += f"\n{Fore.MAGENTA}Give conditions on when can the rule be evaluated for:"
        case_query.scope.update({'case': case_query.case})
        shell = None
        if self.viewer is None:
            prompt_str = self.construct_prompt_str_for_shell(case_query, prompt_for, prompt_str)
            shell = IPythonShell(header=prompt_str, prompt_for=prompt_for, case_query=case_query,
                                    code_to_modify=code_to_modify)
        else:

            self.viewer.update_for_case_query(case_query, prompt_str,
                                                prompt_for=prompt_for, code_to_modify=code_to_modify)
        user_input, expression_tree = self.prompt_user_input_and_parse_to_expression(shell=shell)
        self.print_func("Exited shell")
        return user_input, expression_tree



    def construct_prompt_str_for_shell(self, case_query: CaseQuery, prompt_for: PromptFor,
                                       prompt_str: Optional[str] = None) -> str:
        """
        Construct the prompt string for the shell.

        :param case_query: The case query to prompt the user for.
        :param prompt_for: The type of information the user should provide for the given case.
        :param prompt_str: The prompt string to display to the user.
        """
        prompt_str += (f"\n{Fore.CYAN}{case_query.name}{Fore.MAGENTA} of type(s) "
                       f"{Fore.CYAN}({', '.join(map(lambda x: x.__name__, case_query.core_attribute_type))}){Fore.MAGENTA}")
        if prompt_for == PromptFor.Conditions:
            prompt_str += (f"\ne.g. `{Fore.GREEN}return {Fore.BLUE}len{Fore.RESET}(case.attribute) > {Fore.BLUE}0` "
                           f"{Fore.MAGENTA}\nOR `{Fore.GREEN}return {Fore.YELLOW}True`{Fore.MAGENTA} (If you want the"
                           f" rule to be always evaluated) \n"
                           f"You can also do {Fore.YELLOW}%edit{Fore.MAGENTA} for more complex conditions.\n")

        prompt_str = f"{Fore.MAGENTA}{prompt_str}{Fore.YELLOW}\n(Write %help for guide){Fore.RESET}\n"
        return prompt_str


    def prompt_user_input_and_parse_to_expression(self, shell: Optional[IPythonShell] = None,
                                                  user_input: Optional[str] = None) \
            -> Tuple[Optional[str], Optional[ast.AST]]:
        """
        Prompt the user for input.

        :param shell: The Ipython shell to use for prompting the user.
        :param user_input: The user input to use. If given, the user input will be used instead of prompting the user.
        :return: The user input and the AST tree.
        """
        while True:
            if user_input is None:
                user_input = self.start_shell_and_get_user_input(shell=shell)
                if user_input is None or user_input == 'exit':
                    return user_input, None
                self.print_func(f"{Fore.GREEN}Captured User input: {Style.RESET_ALL}")
                highlighted_code = highlight(user_input, PythonLexer(), TerminalFormatter())
                self.print_func(highlighted_code)
            try:
                return user_input, parse_string_to_expression(user_input)
            except Exception as e:
                msg = f"Error parsing expression: {e}"
                logging.error(msg)
                self.print_func(f"{Fore.RED}{msg}{Style.RESET_ALL}")
                user_input = None

    def start_shell_and_get_user_input(self, shell: Optional[IPythonShell] = None) -> Optional[str]:
        """
        Start the shell and get user input.

        :param shell: The Ipython shell to use for prompting the user.
        :return: The user input.
        """
        if self.viewer is None:
            shell = IPythonShell() if shell is None else shell
            shell.run()
            user_input = shell.user_input
        else:
            app = QApplication.instance()
            if app is None:
                raise RuntimeError("QApplication instance is None. Please run the application first.")
            self.viewer.show()
            app.exec()
            user_input = self.viewer.user_input
        return user_input



import http.server
import socketserver
import threading
import time
import os
import sys

dot_path = None
last_modified = None

def get_dot_content():
    with open(dot_path, "r", encoding="utf-8") as f:
        return f.read().replace("`", "\\`")

def generate_html(dot_output="graph.html"):
    html_template = f"""<!DOCTYPE html>
<html>
<head>
  <meta charset="UTF-8">
  <title>Live DOT Graph</title>
  <script src="https://unpkg.com/viz.js@2.1.2/viz.js"></script>
  <script src="https://unpkg.com/viz.js@2.1.2/full.render.js"></script>
</head>
<body>
  <h2>Live DOT Graph</h2>
  <div id="graph">Loading...</div>
  <script>
    const viz = new Viz();

    async function fetchAndRender() {{
      try {{
        const res = await fetch('/graph.dot?' + Date.now());
        const dot = await res.text();
        const svg = await viz.renderSVGElement(dot);
        const container = document.getElementById('graph');
        container.innerHTML = '';
        container.appendChild(svg);
      }} catch (err) {{
        document.getElementById('graph').innerHTML = '<pre>' + err + '</pre>';
      }}
    }}

    fetchAndRender();
    setInterval(fetchAndRender, 2000);  // Refresh every 2 seconds
  </script>
</body>
</html>"""
    with open(dot_output, "w", encoding="utf-8") as f:
        f.write(html_template)

class CustomHandler(http.server.SimpleHTTPRequestHandler):
    def do_GET(self):
        if self.path.startswith("/graph.dot"):
            self.send_response(200)
            self.send_header("Content-type", "text/plain")
            self.end_headers()
            self.wfile.write(get_dot_content().encode("utf-8"))
        else:
            super().do_GET()

def serve_dot(dot_file, port=8000):
    global dot_path
    dot_path = dot_file
    while not os.path.exists(dot_path):
        print(f"Waiting for {dot_path} to be created...")
        time.sleep(1)
    generate_html("graph.html")

    handler = CustomHandler
    with socketserver.TCPServer(("", port), handler) as httpd:
        print(f"Serving at http://localhost:{port}/graph.html")
        try:
            httpd.serve_forever()
        except KeyboardInterrupt:
            print("\nServer stopped.")

if __name__ == "__main__":
    if len(sys.argv) < 2:
        print("Usage: python live_dot_server.py path/to/graph.dot [port]")
        sys.exit(1)

    dot_file = sys.argv[1]
    port = int(sys.argv[2]) if len(sys.argv) > 2 else 8000

    if not os.path.exists(dot_file):
        print(f"Error: File '{dot_file}' does not exist.")
        sys.exit(1)

    serve_dot(dot_file, port)



from __future__ import annotations

import os
from unittest import TestCase

from typing_extensions import List, Any

from .datasets import Robot, Part, PhysicalObject
from ripple_down_rules.datastructures.case import CaseAttribute
from ripple_down_rules.datastructures.dataclasses import CaseQuery, CallableExpression
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import SingleClassRDR


class RelationalRDRTestCase(TestCase):
    case: Any
    case_query: Any
    test_results_dir: str = "./test_results"
    expert_answers_dir: str = "./test_expert_answers"
    robot: Robot
    part_a: Part
    part_b: Part
    part_c: Part
    part_d: Part
    part_e: Part
    part_f: Part
    target: List[PhysicalObject]

    @classmethod
    def setUpClass(cls):
        if not os.path.exists(cls.test_results_dir):
            os.makedirs(cls.test_results_dir)
        cls.part_a = Part(name="A")
        cls.part_b = Part(name="B")
        cls.part_c = Part(name="C")
        cls.part_d = Part(name="D")
        cls.part_e = Part(name="E")
        cls.part_f = Part(name="F")
        robot = Robot("pr2", parts=[cls.part_a, cls.part_b, cls.part_c, cls.part_d])
        cls.part_a.contained_objects = [cls.part_b, cls.part_c]
        cls.part_c.contained_objects = [cls.part_d]
        cls.part_d.contained_objects = [cls.part_e]
        cls.part_e.contained_objects = [cls.part_f]
        cls.robot: Robot = robot
        cls.case_query = CaseQuery(robot, "contained_objects", (PhysicalObject,), False)
        cls.target = [cls.part_b, cls.part_c, cls.part_d, cls.part_e]

    def test_classify_scrdr(self):
        use_loaded_answers = True
        save_answers = False
        filename = self.expert_answers_dir + "/relational_scrdr_expert_answers_classify"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)

        scrdr = SingleClassRDR()
        cat = scrdr.fit_case(CaseQuery(self.robot, "contained_objects", (PhysicalObject,), False), expert=expert)
        # render_tree(scrdr.start_rule, use_dot_exporter=True,
        #             filename=self.test_results_dir + "/relational_scrdr_classify")
        self.assertEqual(cat, self.target)

        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def test_parse_relational_conditions(self):
        user_input = "case.parts is not None and len(case.parts) > 0"
        conditions = CallableExpression(user_input, bool)
        self.assertEqual(conditions(self.robot), (self.robot.parts is not None and len(self.robot.parts) > 0))

    def test_parse_relational_conclusions(self):
        user_input = "case.parts.contained_objects"
        conclusion = CallableExpression(user_input, (CaseAttribute, PhysicalObject,),
        mutually_exclusive=False)
        self.assertEqual(conclusion(self.robot), self.target)



from __future__ import annotations

from collections import UserDict
from copy import copy, deepcopy
from dataclasses import dataclass, is_dataclass
from enum import Enum

from pandas import DataFrame
from sqlalchemy import MetaData
from sqlalchemy.orm import DeclarativeBase as SQLTable, registry
from typing_extensions import Any, Optional, Dict, Type, Set, Hashable, Union, List, TYPE_CHECKING

from ..utils import make_set, row_to_dict, table_rows_as_str, get_value_type_from_type_hint, SubclassJSONSerializer, \
    get_full_class_name, get_type_from_string, make_list, is_iterable, serialize_dataclass, dataclass_to_dict, copy_case

if TYPE_CHECKING:
    from ripple_down_rules.rules import Rule
    from .callable_expression import CallableExpression


class Case(UserDict, SubclassJSONSerializer):
    """
    A collection of attributes that represents a set of constraints on a case. This is a dictionary where the keys are
    the names of the attributes and the values are the attributes. All are stored in lower case.
    """

    def __init__(self, _obj_type: Type, _id: Optional[Hashable] = None,
                 _name: Optional[str] = None, original_object: Optional[Any] = None, **kwargs):
        """
        Create a new row.

        :param _obj_type: The type of the object that the row represents.
        :param _id: The id of the row.
        :param _name: The semantic name that describes the row.
        :param kwargs: The attributes of the row.
        """
        super().__init__(kwargs)
        self._original_object = original_object
        self._obj_type: Type = _obj_type
        self._id: Hashable = _id if _id is not None else id(self)
        self._name: str = _name if _name is not None else self._obj_type.__name__

    @classmethod
    def from_obj(cls, obj: Any, obj_name: Optional[str] = None, max_recursion_idx: int = 3) -> Case:
        """
        Create a row from an object.

        :param obj: The object to create a row from.
        :param max_recursion_idx: The maximum recursion index to prevent infinite recursion.
        :param obj_name: The name of the object.
        :return: The row of the object.
        """
        return create_case(obj, max_recursion_idx=max_recursion_idx, obj_name=obj_name)

    def __getitem__(self, item: str) -> Any:
        return super().__getitem__(item.lower())

    def __setitem__(self, name: str, value: Any):
        name = name.lower()
        if name in self:
            if isinstance(self[name], list):
                self[name].extend(make_list(value))
            elif isinstance(value, list):
                new_list = make_list(self[name])
                new_list.extend(make_list(value))
                super().__setitem__(name, new_list)
            else:
                super().__setitem__(name, value)
        else:
            super().__setitem__(name, value)
        setattr(self, name, self[name])

    def __contains__(self, item):
        if isinstance(item, (type, Enum)):
            item = item.__name__
        return super().__contains__(item.lower())

    def __delitem__(self, key):
        super().__delitem__(key.lower())

    def __hash__(self):
        return self._id

    def _to_json(self) -> Dict[str, Any]:
        serializable = {k: v for k, v in self.items() if not k.startswith("_")}
        serializable["_id"] = self._id
        serializable["_obj_type"] = get_full_class_name(self._obj_type) if self._obj_type is not None else None
        serializable["_name"] = self._name
        for k, v in serializable.items():
            if isinstance(v, set):
                serializable[k] = {'_type': get_full_class_name(set), 'value': serialize_dataclass(list(v))}
            else:
                serializable[k] = serialize_dataclass(v)
        return {k: v.to_json() if isinstance(v, SubclassJSONSerializer) else v for k, v in serializable.items()}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Case:
        id_ = data.pop("_id")
        obj_type = get_type_from_string(data.pop("_obj_type")) if data["_obj_type"] is not None else None
        name = data.pop("_name")
        for k, v in data.items():
            data[k] = SubclassJSONSerializer.from_json(v)
        return cls(_obj_type=obj_type, _id=id_, _name=name, **data)

    def __deepcopy__(self, memo: Dict[Hashable, Any]) -> Case:
        """
        Create a deep copy of the case.

        :param memo: A dictionary to keep track of objects that have already been copied.
        :return: A deep copy of the case.
        """
        new_case = Case(self._obj_type, _id=self._id, _name=self._name, original_object=self._original_object)
        for k, v in self.items():
            new_case[k] = deepcopy(v)
        return new_case

    def __copy__(self) -> Case:
        """
        Create a shallow copy of the case.

        :return: A shallow copy of the case.
        """
        new_case = Case(self._obj_type, _id=self._id, _name=self._name, original_object=self._original_object)
        for k, v in self.items():
            new_case[k] = copy(v)
        return new_case


@dataclass
class CaseAttributeValue(SubclassJSONSerializer):
    """
    A column value is a value in a column.
    """
    id: Hashable
    """
    The row id of the column value.
    """
    value: Any
    """
    The value of the column.
    """

    def __eq__(self, other):
        if not isinstance(other, CaseAttributeValue):
            return False
        return self.value == other.value

    def __hash__(self):
        return self.id

    def _to_json(self) -> Dict[str, Any]:
        return {"id": self.id, "value": self.value}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> CaseAttributeValue:
        return cls(id=data["id"], value=data["value"])


class CaseAttribute(list, SubclassJSONSerializer):
    nullable: bool = True
    """
    A boolean indicating whether the case attribute can be None or not.
    """
    mutually_exclusive: bool = False
    """
    A boolean indicating whether the case attribute is mutually exclusive or not. (i.e. can only have one value)
    """

    @classmethod
    def from_obj(cls, values: List[Any]) -> CaseAttribute:
        return cls(make_list(values))

    @property
    def as_dict(self) -> Dict[str, Any]:
        """
        Get the case attribute as a dictionary.

        :return: The case attribute as a dictionary.
        """
        return {self.__class__.__name__: self}

    def filter_by(self, condition: CallableExpression) -> CaseAttribute:
        """
        Filter the column by a condition.

        :param condition: The condition to filter by.
        :return: The filtered column.
        """
        return self.__class__([v for v in self if condition(v)])

    def __eq__(self, other):
        if not isinstance(other, list):
            return super().__eq__(make_list(other))
        return super().__eq__(other)

    def __hash__(self):
        return hash(id(self))

    def __str__(self):
        if len(self) == 0:
            return "None"
        return str([v for v in self]) if len(self) > 1 else str(next(iter(self)))

    def _to_json(self) -> Dict[str, Any]:
        return {str(i): v.to_json() if isinstance(v, SubclassJSONSerializer) else v
                for i, v in enumerate(self)}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> CaseAttribute:
        return cls([SubclassJSONSerializer.from_json(v) for _, v in data.items()])


def create_cases_from_dataframe(df: DataFrame, name: Optional[str] = None) -> List[Case]:
    """
    Create cases from a pandas DataFrame.

    :param df: The DataFrame to create cases from.
    :param name: The semantic name of the DataFrame that describes the DataFrame.
    :return: The cases of the DataFrame.
    """
    cases = []
    attribute_names = list(df.columns)
    for row_id, case in df.iterrows():
        case = {col_name: case[col_name].item() for col_name in attribute_names}
        cases.append(Case(DataFrame, _id=row_id, _name=name, **case))
    return cases


def create_case(obj: Any, recursion_idx: int = 0, max_recursion_idx: int = 0,
                obj_name: Optional[str] = None, parent_is_iterable: bool = False) -> Case:
    """
    Create a case from an object.

    :param obj: The object to create a case from.
    :param recursion_idx: The current recursion index.
    :param max_recursion_idx: The maximum recursion index to prevent infinite recursion.
    :param obj_name: The name of the object.
    :param parent_is_iterable: Boolean indicating whether the parent object is iterable or not.
    :return: The case that represents the object.
    """
    obj_name = obj_name or obj.__class__.__name__
    if isinstance(obj, DataFrame):
        return create_cases_from_dataframe(obj, obj_name)
    if isinstance(obj, Case) or (is_dataclass(obj) and not isinstance(obj, SQLTable)):
        return obj
    if ((recursion_idx > max_recursion_idx)
            or (obj.__class__.__module__ == "builtins" and not isinstance(obj, (list, set, dict)))
            or (obj.__class__ in [MetaData, registry])):
        return Case(type(obj), _id=id(obj), _name=obj_name, original_object=obj,
                    **{obj_name or obj.__class__.__name__: make_list(obj) if parent_is_iterable else obj})
    case = Case(type(obj), _id=id(obj), _name=obj_name, original_object=obj)
    if isinstance(obj, dict):
        for k, v in obj.items():
            case = create_or_update_case_from_attribute(v, k, obj, obj_name, recursion_idx,
                                                        max_recursion_idx, parent_is_iterable, case)
    for attr in dir(obj):
        if attr.startswith("_") or callable(getattr(obj, attr)):
            continue
        attr_value = getattr(obj, attr)
        case = create_or_update_case_from_attribute(attr_value, attr, obj, obj_name, recursion_idx,
                                                    max_recursion_idx, parent_is_iterable, case)
    return case


def create_or_update_case_from_attribute(attr_value: Any, name: str, obj: Any, obj_name: Optional[str] = None,
                                         recursion_idx: int = 0, max_recursion_idx: int = 1,
                                         parent_is_iterable: bool = False,
                                         case: Optional[Case] = None) -> Case:
    """
    Create or update a case from an attribute of the object that the case represents.

    :param attr_value: The attribute value.
    :param name: The name of the attribute.
    :param obj: The parent object of the attribute.
    :param obj_name: The parent object name.
    :param recursion_idx: The recursion index to prevent infinite recursion.
    :param max_recursion_idx: The maximum recursion index.
    :param parent_is_iterable: Boolean indicating whether the parent object is iterable or not.
    :param case: The case to update.
    :return: The updated/created case.
    """
    if case is None:
        case = Case(type(obj), _id=id(obj), _name=obj_name, original_object=obj)
    if isinstance(attr_value, (dict, UserDict)):
        case.update({f"{obj_name}.{k}": v for k, v in attr_value.items()})
    if hasattr(attr_value, "__iter__") and not isinstance(attr_value, str):
        column = create_case_attribute_from_iterable_attribute(attr_value, name, obj, obj_name,
                                                               recursion_idx=recursion_idx + 1,
                                                               max_recursion_idx=max_recursion_idx)
        case[name] = column
    else:
        case[name] = make_list(attr_value) if parent_is_iterable else attr_value
    return case


def create_case_attribute_from_iterable_attribute(attr_value: Any, name: str, obj: Any, obj_name: Optional[str] = None,
                                                  recursion_idx: int = 0,
                                                  max_recursion_idx: int = 1) -> CaseAttribute:
    """
    Get a case attribute from an iterable attribute.

    :param attr_value: The iterable attribute to get the case from.
    :param name: The name of the case.
    :param obj: The parent object of the iterable.
    :param obj_name: The parent object name.
    :param recursion_idx: The recursion index to prevent infinite recursion.
    :param max_recursion_idx: The maximum recursion index.
    :return: A case attribute that represents the original iterable attribute.
    """
    values = list(attr_value.values()) if isinstance(attr_value, (dict, UserDict)) else attr_value
    try:
        _type = type(list(values)[0]) if len(values) > 0 else get_value_type_from_type_hint(name, obj)
    except ValueError:
        _type = None
    attr_case = Case(_type, _id=id(attr_value), _name=name, original_object=attr_value)
    case_attr = CaseAttribute(values)
    for idx, val in enumerate(values):
        sub_attr_case = create_case(val, recursion_idx=recursion_idx,
                                    max_recursion_idx=max_recursion_idx,
                                    obj_name=name, parent_is_iterable=True)
        attr_case.update(sub_attr_case)
    for sub_attr, val in attr_case.items():
        try:
            setattr(case_attr, sub_attr, val)
        except AttributeError:
            pass
    return case_attr


def show_current_and_corner_cases(case: Any, targets: Optional[Dict[str, Any]] = None,
                                  current_conclusions: Optional[Dict[str, Any]] = None,
                                  last_evaluated_rule: Optional[Rule] = None) -> str:
    """
    Get the the data to show of the new case and if last evaluated rule exists also show that of the corner case.

    :param case: The new case.
    :param targets: The target attribute of the case.
    :param current_conclusions: The current conclusions of the case.
    :param last_evaluated_rule: The last evaluated rule in the RDR.
    """
    corner_case = None
    targets = {f"target_{name}": value for name, value in targets.items()} if targets else {}
    current_conclusions = {name: value for name, value in current_conclusions.items()} if current_conclusions else {}
    information = ""
    if last_evaluated_rule:
        action = "Refinement" if last_evaluated_rule.fired else "Alternative"
        information += f"{action} needed for rule: {last_evaluated_rule}\n"
        corner_case = last_evaluated_rule.corner_case

    corner_row_dict = None
    if isinstance(case, SQLTable):
        case_dict = row_to_dict(case)
        if last_evaluated_rule and last_evaluated_rule.fired:
            corner_row_dict = row_to_dict(last_evaluated_rule.corner_case)
    elif is_dataclass(case):
        case_dict = dataclass_to_dict(case)
        if last_evaluated_rule and last_evaluated_rule.fired:
            corner_row_dict = dataclass_to_dict(last_evaluated_rule.corner_case)
    else:
        case_dict = copy_case(case)
        if last_evaluated_rule and last_evaluated_rule.fired:
            corner_row_dict = copy_case(corner_case)

    case_dict.update(targets)
    case_dict.update(current_conclusions)
    all_table_rows = [case_dict]
    if corner_row_dict:
        corner_conclusion = last_evaluated_rule.conclusion(case)
        corner_row_dict.update({corner_conclusion.__class__.__name__: corner_conclusion})
        all_table_rows.append(corner_row_dict)
    information += "\n" + "=" * 50 + "\n"
    information += "\n" + table_rows_as_str(all_table_rows) + "\n"
    return information
